""" Link building classes for Frame Graph

The classes below are the various instances used by the Scope.resolve_links
factory. After the pose scaffolding has been built (Scope.build_scaffold) the
classes below can be used to retrieve the links that will connect the frames of
the final frame graph.

"""


from typing import Union, Tuple
import numpy as np
from scipy.spatial.transform import Rotation as ScipyRotation

from .... import transform as tf
from .scopes import Scope, SdfLink


class CustomLink(SdfLink):
    def __init__(
        self, parent: Union[str, tf.Frame], child: Union[str, tf.Frame], link: tf.Link
    ) -> None:
        super().__init__(parent, child)
        self._link = link

    def to_transform_link(self, scope: "Scope", shape: tuple, axis: int) -> tf.Link:
        return self._link


class SimplePose(SdfLink):
    """A fixed pose that doesn't need the scaffold"""

    def __init__(
        self, parent: Union[str, tf.Frame], child: Union[str, tf.Frame], pose: str
    ) -> None:
        super().__init__(parent, child)
        self.pose = np.array(pose.split(), dtype=float)

    def to_transform_link(self, scope: "Scope", shape: tuple, axis: int) -> tf.Link:
        offset = np.broadcast_to(self.pose[:3], shape)
        angles = np.broadcast_to(self.pose[3:], shape)

        return tf.CompundLink(
            [
                tf.Translation(-offset, axis=axis),
                tf.EulerRotation("xyz", -angles, axis=axis),
            ]
        )


class DynamicPose(SdfLink):
    """A pose that needs to be worked out after scaffolding has been built."""

    def __init__(
        self,
        parent: Union[str, tf.Frame],
        child: Union[str, tf.Frame],
        *,
        scaffold_parent: Union[str, tf.Frame] = None,
        scaffold_child: Union[str, tf.Frame] = None
    ) -> None:
        super().__init__(parent, child)

        self.scaffold_parent = scaffold_parent
        if scaffold_parent is None:
            self.scaffold_parent = parent

        self.scaffold_child = scaffold_child
        if scaffold_child is None:
            self.scaffold_child = child

    def to_transform_link(self, scope: "Scope", shape: tuple, axis: int) -> tf.Link:
        parent = self.scaffold_parent
        child = self.scaffold_child

        if isinstance(parent, str):
            parent = scope.get(parent, scaffolding=True)

        if isinstance(child, str):
            child = scope.get(child, scaffolding=True)

        translation = parent.transform((0, 0, 0), child)

        # link's euler angles
        rot_matrix = list()
        for basis in np.eye(3):
            basis = parent.transform(basis, child)
            rot_matrix.append(basis)
        rot_matrix = np.stack(rot_matrix, axis=1)
        rot_matrix -= translation[:, None]
        angles = ScipyRotation.from_matrix(rot_matrix).as_euler("xyz")

        translation = np.broadcast_to(translation, shape)
        angles = np.broadcast_to(angles, shape)

        return tf.CompundLink(
            [
                tf.Translation(-translation, axis=axis),
                tf.EulerRotation("xyz", -angles, axis=axis),
            ]
        )


class SingleAxisJoint(SdfLink):
    """Base class for joints using an axis"""

    def __init__(
        self,
        parent: Union[str, tf.Frame],
        child: Union[str, tf.Frame],
        axis: str,
        expressed_in: str,
        lower_limit: float,
        upper_limit: float,
    ) -> None:
        super().__init__(parent, child)
        self.axis = np.array(axis.split(" "), dtype=float)
        self.expressed_in = parent
        self.expressed_in: Union[str, tf.Frame] = expressed_in
        self.lower_limit = lower_limit
        self.upper_limit = upper_limit

    def _convert_axis(self, scope: "Scope", shape: Tuple[int]) -> np.ndarray:
        expressed_in = scope.get(self.expressed_in, scaffolding=True)
        parent = scope.get(self.parent, scaffolding=True)

        rotvec = expressed_in.transform(self.axis, parent)
        rotvec /= np.linalg.norm(rotvec)
        rotvec = np.broadcast_to(rotvec, shape)

        return rotvec


class RotationJoint(SingleAxisJoint):
    def to_transform_link(self, scope: "Scope", shape: tuple, axis: int) -> tf.Link:
        rotvec = self._convert_axis(scope, shape)
        angle_shape = [*shape]
        angle_shape.pop(axis)
        angle = np.zeros(angle_shape)

        # TODO: remove clip in favor of SDF state
        angle = np.clip(angle, self.lower_limit, self.upper_limit)

        return tf.RotationalJoint(
            rotvec,
            angle=angle,
            axis=axis,
            lower_limit=self.lower_limit,
            upper_limit=self.upper_limit,
        )


class PrismaticJoint(SingleAxisJoint):
    def to_transform_link(self, scope: "Scope", shape: tuple, axis: int) -> tf.Link:
        direction = self._convert_axis(scope, shape)
        amount_shape = [*shape]
        amount_shape.pop(axis)
        amount = np.ones(amount_shape)

        # TODO: remove clip in favor of SDF state
        amount = np.clip(amount, self.lower_limit, self.upper_limit)

        return tf.PrismaticJoint(
            direction,
            amount=np.ones(amount_shape),
            axis=axis,
            lower_limit=self.lower_limit,
            upper_limit=self.upper_limit,
        )



from typing import List, Union

from .scopes import Scope
from .generic import (
    GenericFrame,
    GenericInclude,
    GenericJoint,
    GenericLink,
    GenericModel,
    GenericPose,
    GenericSensor,
    GenericLight,
    GenericWorld,
    NamedPoseBearing,
    PoseBearing,
)
from .factory import FactoryBase
from .. import sdformat
from ..bindings import v18


IncludeElement = Union[v18.ModelModel.Include, v18.World.Include]
FrameElement = Union[v18.ModelModel.Frame, v18.World.Frame]


class Converter(FactoryBase):
    """Functions to convert v1.8 SDF objects into generic SDF objects."""

    def __call__(self, sdf: str) -> Union[Scope, List[Scope]]:
        """Convert v1.8 SDF into a Scope

        Parameters
        ----------
        sdf : str
            A string containing v1.8 SDFormat XML.

        Returns
        -------
        graph : _graph.Graph
            A container storing the nodes and edges of the frame graph.

        links : Dict[str, Frames]
            A dict of (named) links in the graph.

        Notes
        -----
        It is necessary to split the parsers, since versions differ so heavily.
        """

        sdf_root: v18.Sdf = sdformat.loads(sdf)
        graph_list: List[Scope] = list()

        for world in sdf_root.world:
            generic_world = self._to_generic_world(world)
            graph = self.convert_world(generic_world)
            graph_list.append(graph)

        if sdf_root.model is not None:
            generic_model = self._to_generic_model(sdf_root.model)
            graph = self.convert_model(generic_model)
            graph_list.append(graph)

        if sdf_root.light is not None:
            generic_light = self._to_generic_light(sdf_root.light)
            graph = self.convert_light(generic_light)
            graph_list.append(graph)

        if sdf_root.actor is not None:
            raise NotImplementedError("Actors are not implemented yet.")

        return graph_list

    def _to_generic_joint(self, joint: v18.Joint) -> GenericJoint:
        sensors = list()
        for sensor in joint.sensor:
            sensors.append(self._to_generic_sensor(sensor))

        joint_args = {
            "name": joint.name,
            "kind": joint.type,
            "parent": joint.parent,
            "child": joint.child,
            "pose": None,
            "sensor": sensors,
        }

        if joint.pose is not None:
            joint_args["pose"] = joint.pose

        if joint.axis is not None:
            axis = GenericJoint.Axis()

            if joint.axis.xyz is not None:
                axis.xyz.value = joint.axis.xyz.value
                axis.xyz.expressed_in = joint.axis.xyz.expressed_in

            if joint.axis.limit is not None:
                axis.limit.lower = joint.axis.limit.lower
                axis.limit.upper = joint.axis.limit.upper
                axis.limit.effort = joint.axis.limit.effort
                axis.limit.velocity = joint.axis.limit.velocity
                axis.limit.stiffness = joint.axis.limit.stiffness
                axis.limit.dissipation = joint.axis.limit.dissipation

            joint_args["axis"] = axis

        return GenericJoint(**joint_args)

    def _to_generic_sensor(self, sensor: v18.Sensor) -> GenericSensor:
        sensor_args = {"name": sensor.name, "type": sensor.type, "pose": sensor.pose}

        if sensor.camera is not None:
            if sensor.camera.noise is not None:
                raise NotImplementedError()
            if sensor.camera.distortion is not None:
                raise NotImplementedError()
            if sensor.camera.lens is not None:
                raise NotImplementedError()

        if sensor.camera is not None:
            image = GenericSensor.Camera.Image()
            if sensor.camera.image is not None:
                image.width = sensor.camera.image.width
                image.height = sensor.camera.image.height
                image.format = sensor.camera.image.format

            sensor_args["camera"] = GenericSensor.Camera(
                name=sensor.camera.name,
                pose=sensor.camera.pose,
                horizontal_fov=sensor.camera.horizontal_fov,
                image=image,
            )

        return GenericSensor(**sensor_args)

    def _to_generic_light(self, light: v18.Light) -> GenericLight:
        return GenericLight(name=light.name, pose=light.pose)

    def _to_generic_link(self, link: v18.Link) -> GenericLink:
        link_args = {
            "name": link.name,
            "must_be_base_link": link.must_be_base_link,
            "pose": link.pose,
            "inertial": None,
            "projector": None,
            "sensors": [self._to_generic_sensor(sensor) for sensor in link.sensor],
            "lights": [self._to_generic_light(light) for light in link.light],
        }

        if link.inertial is not None:
            link_args["inertial"] = GenericLink.Inertial(
                pose=GenericPose(value=link.inertial.pose, relative_to=link.name)
            )

        link_args["collisions"] = [
            NamedPoseBearing(name=c.name, pose=c.pose)
            if c.pose is not None
            else NamedPoseBearing(name=c.name)
            for c in link.collision
        ]

        link_args["visuals"] = [
            NamedPoseBearing(name=v.name, pose=v.pose)
            if v.pose is not None
            else NamedPoseBearing(name=v.name)
            for v in link.visual
        ]

        if link.projector is not None:
            link_args["projector"] = NamedPoseBearing(
                name=link.projector.name, pose=link.projector.pose
            )

        link_args["audio_source_poses"] = [
            a.pose if a.pose is not None else GenericPose() for a in link.audio_source
        ]

        return GenericLink(**link_args)

    def _to_generic_model(self, model: v18.ModelModel) -> GenericModel:
        if len(model.gripper) > 0:
            raise NotImplementedError(
                "Gripper not implemented yet (lacking upstream docs)."
            )

        return GenericModel(
            name=model.name,
            pose=model.pose,
            placement_frame=model.placement_frame,
            canonical_link=model.canonical_link,
            links=[self._to_generic_link(l) for l in model.link],
            include=[self._to_generic_include(i) for i in model.include],
            models=[self._to_generic_model(m) for m in model.model],
            joints=[self._to_generic_joint(j) for j in model.joint],
            frames=[self._to_generic_frame(f) for f in model.frame],
        )

    def _to_generic_frame(self, frame: FrameElement) -> GenericFrame:
        return GenericFrame(
            attached_to=frame.attached_to, name=frame.name, pose=frame.pose
        )

    def _to_generic_include(self, include: IncludeElement) -> GenericInclude:
        return GenericInclude(
            name=include.name,
            pose=include.pose,
            placement_frame=include.placement_frame,
            uri=include.uri,
        )

    def _to_generic_population(
        self, population: v18.World.Population
    ) -> GenericWorld.GenericPopulation:
        distribution = GenericWorld.GenericPopulation.GenericDistribution()
        if population.distribution is not None:
            distribution.type = population.distribution.type
            distribution.step = population.distribution.step
            distribution.cols = population.distribution.cols
            distribution.rows = population.distribution.rows

        return GenericWorld.GenericPopulation(
            name=population.name,
            pose=population.pose,
            model_count=population.model_count,
            distribution=distribution,
            box=population.box,
            cylinder=population.cylinder,
            model=self._to_generic_model(population.model),
        )

    def _to_generic_world(self, world: v18.World) -> GenericWorld:
        return GenericWorld(
            name=world.name,
            includes=[self._to_generic_include(i) for i in world.include],
            models=[self._to_generic_model(m) for m in world.model],
            frames=[self._to_generic_frame(f) for f in world.frame],
            lights=[self._to_generic_light(l) for l in world.light],
            population=[self._to_generic_population(p) for p in world.population],
        )



from dataclasses import dataclass, field
from typing import List, Optional

__NAMESPACE__ = "sdformat/v1.2/geometry.xsd"


@dataclass
class Geometry:
    """
    The shape of the visual or collision object.

    Parameters
    ----------
    box: Box shape
    sphere: Sphere shape
    cylinder: Cylinder shape
    mesh: Mesh shape
    plane: Plane shape
    image: Extrude a set of boxes from a grayscale image.
    heightmap: A heightmap based on a 2d grayscale image.
    """

    class Meta:
        name = "geometry"

    box: Optional["Geometry.Box"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    sphere: Optional["Geometry.Sphere"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    cylinder: Optional["Geometry.Cylinder"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    mesh: Optional["Geometry.Mesh"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    plane: Optional["Geometry.Plane"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    image: Optional["Geometry.Image"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    heightmap: Optional["Geometry.Heightmap"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )

    @dataclass
    class Box:
        """
        Box shape.

        Parameters
        ----------
        size: The three side lengths of the box. The origin of the box
            is in its geometric center (inside the center of the box).
        """

        size: str = field(
            default="1 1 1",
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
                "pattern": r"(\s*(-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){2}((-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+))\s*",
            },
        )

    @dataclass
    class Sphere:
        """
        Sphere shape.

        Parameters
        ----------
        radius: radius of the sphere
        """

        radius: float = field(
            default=1.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )

    @dataclass
    class Cylinder:
        """
        Cylinder shape.

        Parameters
        ----------
        radius: Radius of the cylinder
        length: Length of the cylinder
        """

        radius: float = field(
            default=1.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        length: float = field(
            default=1.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )

    @dataclass
    class Mesh:
        """
        Mesh shape.

        Parameters
        ----------
        filename: Mesh filename. DEPRECATED
        uri: Mesh uri
        scale: Scaling factor applied to the mesh
        """

        filename: str = field(
            default="__default__",
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        uri: str = field(
            default="__default__",
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        scale: str = field(
            default="1 1 1",
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
                "pattern": r"(\s*(-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){2}((-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+))\s*",
            },
        )

    @dataclass
    class Plane:
        """
        Plane shape.

        Parameters
        ----------
        normal: Normal direction for the plane
        size: Length of each side of the plane
        """

        normal: str = field(
            default="0 0 1",
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
                "pattern": r"(\s*(-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){2}((-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+))\s*",
            },
        )
        size: str = field(
            default="1 1",
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
                "pattern": r"(\s*(-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+)((-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+))\s*",
            },
        )

    @dataclass
    class Image:
        """
        Extrude a set of boxes from a grayscale image.

        Parameters
        ----------
        uri: URI of the grayscale image file
        scale: Scaling factor applied to the image
        threshold: Grayscale threshold
        height: Height of the extruded boxes
        granularity: The amount of error in the model
        """

        uri: str = field(
            default="__default__",
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        scale: float = field(
            default=1.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        threshold: int = field(
            default=200,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        height: float = field(
            default=1.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        granularity: int = field(
            default=1,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )

    @dataclass
    class Heightmap:
        """
        A heightmap based on a 2d grayscale image.

        Parameters
        ----------
        uri: URI to a grayscale image file
        size: The size of the heightmap in world units
        pos: A position offset.
        texture: The heightmap can contain multiple textures. The order
            of the texture matters. The first texture will appear at the
            lowest height, and the last texture at the highest height.
            Use blend to control the height thresholds and fade between
            textures.
        blend: The blend tag controls how two adjacent textures are
            mixed. The number of blend elements should equal one less
            than the number of textures.
        """

        uri: str = field(
            default="__default__",
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        size: str = field(
            default="1 1 1",
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
                "pattern": r"(\s*(-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){2}((-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+))\s*",
            },
        )
        pos: str = field(
            default="0 0 0",
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
                "pattern": r"(\s*(-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){2}((-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+))\s*",
            },
        )
        texture: List["Geometry.Heightmap.Texture"] = field(
            default_factory=list,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )
        blend: List["Geometry.Heightmap.Blend"] = field(
            default_factory=list,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )

        @dataclass
        class Texture:
            """The heightmap can contain multiple textures.

            The order of the texture matters. The first texture will
            appear at the lowest height, and the last texture at the
            highest height. Use blend to control the height thresholds
            and fade between textures.

            Parameters
            ----------
            size: Size of the applied texture in meters.
            diffuse: Diffuse texture image filename
            normal: Normalmap texture image filename
            """

            size: float = field(
                default=10.0,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            diffuse: str = field(
                default="__default__",
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            normal: str = field(
                default="__default__",
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )

        @dataclass
        class Blend:
            """The blend tag controls how two adjacent textures are mixed.

            The number of blend elements should equal one less than the
            number of textures.

            Parameters
            ----------
            min_height: Min height of a blend layer
            fade_dist: Distance over which the blend occurs
            """

            min_height: float = field(
                default=0.0,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            fade_dist: float = field(
                default=0.0,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )



from typing import List, Union

from .scopes import Scope
from .factory import FactoryBase
from .. import sdformat
from ..bindings import v15
from .generic import (
    GenericFrame,
    GenericInclude,
    GenericLight,
    GenericModel,
    GenericPose,
    GenericSensor,
    GenericJoint,
    GenericWorld,
    NamedPoseBearing,
    GenericLink,
    PoseBearing,
)


IncludeElement = Union[v15.ModelModel.Include, v15.World.Include]
FrameElement = Union[
    v15.ModelModel.Frame, v15.Sensor.Frame, v15.Joint.Frame, v15.Link.Frame
]
PoseElement = Union[
    v15.ModelModel.Frame.Pose,
    v15.Sensor.Frame.Pose,
    v15.Joint.Frame.Pose,
    v15.Link.Frame.Pose,
    v15.Joint.Pose,
]


class Converter(FactoryBase):
    """Functions to convert v1.5 SDF objects into generic SDF objects."""

    def __call__(self, sdf: str) -> Union[Scope, List[Scope]]:
        """Convert v1.5 SDF into a Scope"""

        sdf_root: v15.Sdf = sdformat.loads(sdf)
        graph_list: List[Scope] = list()

        for world in sdf_root.world:
            generic_world = self._to_generic_world(world)
            graph = self.convert_world(generic_world)
            graph_list.append(graph)

        for model in sdf_root.model:
            generic_model = self._to_generic_model(model)
            graph = self.convert_model(generic_model)
            graph_list.append(graph)

        for light in sdf_root.light:
            generic_light = self._to_generic_light(light)
            graph = self.convert_light(generic_light)
            graph_list.append(graph)

        return graph_list

    def _to_generic_pose(self, pose: PoseElement) -> GenericPose:
        if pose is None:
            return GenericPose()
        else:
            return GenericPose(value=pose.value, relative_to=pose.frame)

    def _to_generic_frame(self, frame: FrameElement, attached_to=None) -> GenericFrame:
        return GenericFrame(
            attached_to=attached_to,
            name=frame.name,
            pose=self._to_generic_pose(frame.pose),
        )

    def _to_generic_sensor(self, sensor: v15.Sensor) -> GenericSensor:
        sensor_args = {
            "name": sensor.name,
            "type": sensor.type,
            "pose": self._to_generic_pose(sensor.pose),
            "frames": [
                self._to_generic_frame(x, attached_to=None) for x in sensor.frame
            ],
        }

        if sensor.camera is not None:
            if sensor.camera.noise is not None:
                raise NotImplementedError()
            if sensor.camera.distortion is not None:
                raise NotImplementedError()
            if sensor.camera.lens is not None:
                raise NotImplementedError()

            sensor_args["camera"] = GenericSensor.Camera(
                name=sensor.camera.name,
                pose=self._to_generic_pose(sensor.camera.pose),
                horizontal_fov=sensor.camera.horizontal_fov,
                frames=[
                    self._to_generic_frame(x, attached_to=None)
                    for x in sensor.camera.frame
                ],
            )

            if sensor.camera.image is not None:
                sensor_args["camera"].image = GenericSensor.Camera.Image(
                    width=sensor.camera.image.width,
                    height=sensor.camera.image.height,
                    format=sensor.camera.image.format,
                )

        return GenericSensor(**sensor_args)

    def _to_generic_joint(self, joint: v15.Joint) -> GenericJoint:
        sensors = list()
        for sensor in joint.sensor:
            sensors.append(self._to_generic_sensor(sensor))

        joint_args = {
            "name": joint.name,
            "kind": joint.type,
            "parent": joint.parent,
            "child": joint.child,
            "pose": None,
            "sensor": sensors,
            "frames": [
                self._to_generic_frame(x, attached_to=joint.name) for x in joint.frame
            ],
        }

        if joint.pose is not None:
            joint_args["pose"] = GenericPose(
                value=joint.pose.value, relative_to=joint.pose.frame
            )

        if joint.axis is not None:
            axis = GenericJoint.Axis()

            axis.xyz.value = joint.axis.xyz
            if not joint.axis.use_parent_model_frame:
                axis.xyz.expressed_in = joint.child
            else:
                axis.xyz.expressed_in = joint.parent

            if joint.axis.limit is not None:
                axis.limit.lower = joint.axis.limit.lower
                axis.limit.upper = joint.axis.limit.upper
                axis.limit.effort = joint.axis.limit.effort
                axis.limit.velocity = joint.axis.limit.velocity
                axis.limit.stiffness = joint.axis.limit.stiffness
                axis.limit.dissipation = joint.axis.limit.dissipation

            joint_args["axis"] = axis

        return GenericJoint(**joint_args)

    def _to_generic_light(self, light: v15.Light) -> GenericLight:
        return GenericLight(
            name=light.name,
            frames=[
                self._to_generic_frame(x, attached_to=light.name) for x in light.frame
            ],
            pose=self._to_generic_pose(light.pose),
        )

    def _to_generic_link(self, link: v15.Link) -> GenericLink:
        link_args = {
            "name": link.name,
            "must_be_base_link": link.must_be_base_link,
            "pose": self._to_generic_pose(link.pose),
            "inertial": None,
            "projector": None,
            "sensors": [self._to_generic_sensor(sensor) for sensor in link.sensor],
            "frames": [
                self._to_generic_frame(f, attached_to=link.name) for f in link.frame
            ]
            # lights may not be part of SDF 1.5; tracking issue:
            #
            # "lights": [self._to_generic_light(light) for light in link.light],
        }

        if link.inertial is not None:
            el = GenericLink.Inertial(
                pose=GenericPose(relative_to=link.name),
                frames=[self._to_generic_frame(f) for f in link.inertial.frame],
            )
            if link.inertial.pose is not None:
                el.pose.value = link.inertial.pose.value
                if link.inertial.pose.frame is not None:
                    raise NotImplementedError(
                        "Unsure how to resolve /inertial/pose/@frame."
                    )
            link_args["inertial"] = el

        link_args["collisions"] = [
            NamedPoseBearing(name=c.name, pose=self._to_generic_pose(c.pose))
            if c.pose is not None
            else NamedPoseBearing(name=c.name)
            for c in link.collision
        ]

        link_args["visuals"] = [
            NamedPoseBearing(name=v.name, pose=self._to_generic_pose(v.pose))
            if v.pose is not None
            else NamedPoseBearing(name=v.name)
            for v in link.visual
        ]

        if link.projector is not None:
            link_args["projector"] = NamedPoseBearing(
                name=link.projector.name,
                pose=self._to_generic_pose(link.projector.pose),
            )

        link_args["audio_source_poses"] = [
            self._to_generic_pose(a.pose) if a.pose is not None else GenericPose()
            for a in link.audio_source
        ]

        return GenericLink(**link_args)

    def _to_generic_include(self, include: IncludeElement) -> GenericInclude:
        return GenericInclude(
            name=include.name,
            pose=self._to_generic_pose(include.pose),
            uri=include.uri,
        )

    def _to_generic_model(self, model: v15.ModelModel) -> GenericModel:
        if len(model.gripper) > 0:
            raise NotImplementedError(
                "Gripper not implemented yet (lacking upstream docs)."
            )

        return GenericModel(
            name=model.name,
            include=[self._to_generic_include(i) for i in model.include],
            models=[self._to_generic_model(m) for m in model.model],
            frames=[
                self._to_generic_frame(f, attached_to="__model__") for f in model.frame
            ],
            pose=self._to_generic_pose(model.pose),
            links=[self._to_generic_link(l) for l in model.link],
            joints=[self._to_generic_joint(j) for j in model.joint],
        )

    def _to_generic_population(
        self, population: v15.World.Population
    ) -> GenericWorld.GenericPopulation:
        distribution = GenericWorld.GenericPopulation.GenericDistribution()
        if population.distribution is not None:
            distribution.type = population.distribution.type
            distribution.step = population.distribution.step
            distribution.cols = population.distribution.cols
            distribution.rows = population.distribution.rows

        generic_population = GenericWorld.GenericPopulation(
            name=population.name,
            pose=self._to_generic_pose(population.pose),
            model_count=population.model_count,
            distribution=distribution,
            model=self._to_generic_model(population.model),
            frames=[self._to_generic_frame(x) for x in population.frame],
        )

        if population.box is not None:
            generic_population.box = GenericWorld.GenericPopulation.GenericBox(
                size=population.box.size
            )

        if population.cylinder is not None:
            generic_population.cylinder = (
                GenericWorld.GenericPopulation.GenericCylinder(
                    radius=population.cylinder.radius, length=population.cylinder.length
                )
            )

        return generic_population

    def _to_generic_world(self, world: v15.World) -> GenericWorld:
        return GenericWorld(
            name=world.name,
            includes=[self._to_generic_include(i) for i in world.include],
            models=[self._to_generic_model(m) for m in world.model],
            frames=list(),
            lights=[self._to_generic_light(l) for l in world.light],
            population=[self._to_generic_population(p) for p in world.population],
        )



from skbot.transform import joints
from typing import Callable, Dict, Union, List, Any
import importlib
from urllib.parse import urlparse
from requests.exceptions import RequestException

from .. import sdformat
from .scopes import LightScope, ModelScope, Scope, WorldScope
from ...fuel import get_fuel_model
from .... import transform as tf
from .generic import (
    GenericFrame,
    GenericInclude,
    GenericJoint,
    GenericLight,
    GenericLink,
    GenericModel,
    GenericSensor,
    GenericWorld,
    NamedPoseBearing,
)
from .links import DynamicPose, CustomLink, RotationJoint, PrismaticJoint, SimplePose
from ...transformations import FrustumProjection


# available SDF elements by version
_converter_roots = {
    "1.0": None,
    "1.2": None,
    "1.3": None,
    "1.4": None,
    "1.5": "..v15",
    "1.6": None,
    "1.7": "..v17",
    "1.8": "..v18",
}


class FactoryBase:
    """Frame Graph Factory

    This class constructs a frame graph from a tree of generic SDF objects (see generic.py). It
    is subclassed by the version-specific factories (vXX.py) which implement conversion from
    version-specific SDF objects to generic objects.

    """

    def __init__(self, *, root_uri: str = None):
        self.root_uri: str = root_uri

    def __call__(self, sdf: str) -> Union[Scope, List[Scope]]:
        raise NotImplementedError()

    def _resolve_include(self, uri: str) -> ModelScope:
        # there is only one fuel server
        fuel_server = "fuel.ignitionrobotics.org"
        uri_parts = urlparse(uri)

        if uri_parts.scheme == "https" and uri_parts.netloc == fuel_server:
            root_uri = uri
            rel_path = "model.sdf"
        elif uri_parts.scheme == "model":
            raise NotImplementedError("Unsure how to resolve model://.")
        elif uri_parts.scheme == "":
            root_uri = self.root_uri
            rel_path = uri
        else:
            raise sdformat.ParseError(f"Unknown URI: {uri}")

        try:
            sdf = get_fuel_model(root_uri, file_path=rel_path)
        except RequestException:
            raise sdformat.ParseError(
                f"Unable to read '{str(rel_path)}' from '{str(root_uri)}'"
            )
        return transform_factory(sdf, root_uri=root_uri)[0]

    def _convert_frame(self, scope: "Scope", frame: GenericFrame):
        scope.declare_frame(frame.name)
        scope.add_scaffold(frame.name, frame.pose.value, frame.pose.relative_to)

        if frame.attached_to is None:
            scope.declare_link(DynamicPose(scope.default_frame, frame.name))
        else:
            scope.declare_link(DynamicPose(frame.attached_to, frame.name))

    def convert_sensor(
        self, sensor: GenericSensor, scope: Scope, attached_to: NamedPoseBearing
    ) -> tf.Frame:
        sensor_frame = tf.Frame(3, name=sensor.name)
        sensor_scaffold = tf.Frame(3, name=sensor.name)
        scope.add_scaffold(sensor_scaffold, sensor.pose.value, sensor.pose.relative_to)
        scope.declare_link(
            DynamicPose(attached_to.name, sensor_frame, scaffold_child=sensor_scaffold)
        )

        if sensor.type == "air_pressure":
            raise NotImplementedError()
        elif sensor.type == "altimeter":
            raise NotImplementedError()
        elif sensor.type == "camera":
            pose = sensor.camera.pose
            camera_frame = tf.Frame(3, name="camera-space")
            camera_scaffold = tf.Frame(3, name="camera-space")
            scope.add_scaffold(camera_scaffold, pose.value, pose.relative_to)
            scope.declare_link(
                DynamicPose(
                    sensor_frame,
                    camera_frame,
                    scaffold_parent=sensor_scaffold,
                    scaffold_child=camera_scaffold,
                )
            )
            scope.declare_link(
                CustomLink(
                    camera_frame,
                    tf.Frame(2, name="pixel-space"),
                    FrustumProjection(
                        sensor.camera.horizontal_fov,
                        (sensor.camera.image.height, sensor.camera.image.width),
                    ),
                )
            )

            for frame in sensor.camera.frames:
                self._convert_frame(scope, frame)

        elif sensor.type == "contact":
            raise NotImplementedError()
        elif sensor.type == "depth_camera":
            raise NotImplementedError()
        elif sensor.type == "force_torque":
            raise NotImplementedError()
        elif sensor.type == "gps":
            raise NotImplementedError()
        elif sensor.type == "gpu_lidar":
            raise NotImplementedError()
        elif sensor.type == "gpu_ray":
            raise NotImplementedError()
        elif sensor.type == "imu":
            raise NotImplementedError()
        elif sensor.type == "lidar":
            raise NotImplementedError()
        elif sensor.type == "logica_camera":
            raise NotImplementedError()
        elif sensor.type == "magnetometer":
            raise NotImplementedError()
        elif sensor.type == "multicamera":
            raise NotImplementedError()
        elif sensor.type == "rfid":
            raise NotImplementedError()
        elif sensor.type == "rfidtag":
            raise NotImplementedError()
        elif sensor.type == "rgbd_camera":
            raise NotImplementedError()
        elif sensor.type == "sonar":
            raise NotImplementedError()
        elif sensor.type == "thermal_camera":
            raise NotImplementedError()
        elif sensor.type == "wireless_receiver":
            raise NotImplementedError()
        elif sensor.type == "wireless_transmitter":
            raise NotImplementedError()
        else:
            raise sdformat.ParseError(f"Unkown sensor type: {sensor.type}")

        for frame in sensor.frames:
            self._convert_frame(scope, frame)

        return sensor_scaffold

    def convert_joint(self, joint: GenericJoint, scope: ModelScope):
        scope.declare_frame(joint.name)
        scope.add_scaffold(joint.name, joint.pose.value, joint.pose.relative_to)
        scope.declare_link(DynamicPose(joint.child, joint.name))

        joint_parent = tf.Frame(3, name=joint.name + "_parent")
        scope.declare_link(
            DynamicPose(joint_parent, joint.parent, scaffold_parent=joint.name)
        )

        if joint.type == "revolute":
            scope.declare_link(
                RotationJoint(
                    joint.name,
                    joint_parent,
                    joint.axis.xyz.value,
                    joint.axis.xyz.expressed_in,
                    joint.axis.limit.lower,
                    joint.axis.limit.upper,
                )
            )
        elif joint.type == "hinge":
            raise NotImplementedError("Hinge joints have not been added yet.")
        elif joint.type == "gearbox":
            raise NotImplementedError("Gearbox joints have not been added yet.")
        elif joint.type == "revolute2":
            raise NotImplementedError("Revolute2 type joint is not added yet.")
        elif joint.type == "prismatic":
            scope.declare_link(
                PrismaticJoint(
                    joint.name,
                    joint_parent,
                    joint.axis.xyz.value,
                    joint.axis.xyz.expressed_in,
                    joint.axis.limit.lower,
                    joint.axis.limit.upper,
                )
            )
        elif joint.type == "ball":
            raise NotImplementedError("Ball joints have not been added yet.")
        elif joint.type == "screw":
            raise NotImplementedError("Screw joints have not been added yet.")
        elif joint.type == "universal":
            raise NotImplementedError("Universal joints have not been added yet.")
        elif joint.type == "fixed":
            scope.declare_link(
                DynamicPose(joint.name, joint_parent, scaffold_child=joint.name)
            )
        else:
            raise sdformat.ParseError(f"Unknown Joint type: {joint.type}")

        for sensor in joint.sensor:
            self.convert_sensor(sensor, scope, attached_to=joint)

        for frame in joint.frames:
            self._convert_frame(scope, frame)

        return scope

    def convert_light(self, light: GenericLight, *, scope: Scope = None) -> Scope:
        if scope is None:
            scope = LightScope(light.name)

        for frame in light.frames:
            self._convert_frame(scope, frame)

        scope.declare_frame(light.name)
        scope.add_scaffold(light.name, light.pose.value, light.pose.relative_to)

        return scope

    def convert_link(self, link: GenericLink, scope: Scope) -> Scope:
        if link.must_be_base_link:
            link.pose.relative_to = "world"
            scope.declare_link(DynamicPose("world", link.name))

        scope.declare_frame(link.name)
        scope.add_scaffold(link.name, link.pose.value, link.pose.relative_to)

        if link.inertial:
            scope.declare_link(
                SimplePose(
                    link.name, tf.Frame(3, name="inertial"), link.inertial.pose.value
                )
            )

        for collision in link.collision:
            frame = tf.Frame(3, name=collision.name)
            scaffold_frame = tf.Frame(3, name=collision.name)
            scope.add_scaffold(
                scaffold_frame, collision.pose.value, collision.pose.relative_to
            )
            scope.declare_link(
                DynamicPose(link.name, frame, scaffold_child=scaffold_frame)
            )

        for visual in link.visual:
            frame = tf.Frame(3, name=visual.name)
            scaffold_frame = tf.Frame(3, name=visual.name)
            scope.add_scaffold(
                scaffold_frame, visual.pose.value, visual.pose.relative_to
            )
            scope.declare_link(
                DynamicPose(link.name, frame, scaffold_child=scaffold_frame)
            )

        for sensor in link.sensors:
            self.convert_sensor(sensor, scope, attached_to=link)

        if link.projector:
            frame = tf.Frame(3, name=link.projector.name)
            scaffold_frame = tf.Frame(3, name=link.projector.name)
            scope.add_scaffold(
                scaffold_frame,
                link.projector.pose.value,
                link.projector.pose.relative_to,
            )
            scope.declare_link(
                DynamicPose(link.name, frame, scaffold_child=scaffold_frame)
            )
            # TODO: there might be a link into the projected frame missing
            # here I don't exactly know how projector works and didn't find
            # docs

        for idx, source in enumerate(link.audio_sources):
            name = link.name + f"-audio-source-{idx}"
            scope.declare_frame(name)
            scope.add_scaffold(name, source.pose.value, source.pose.relative_to)
            scope.declare_link(DynamicPose(link.name, name))

        for light in link.lights:
            self.convert_light(light, scope=scope)
            scope.declare_link(DynamicPose(link.name, light.name))

        for frame in link.frames:
            self._convert_frame(scope, frame)

        return scope

    def convert_model(
        self, model: GenericModel, *, parent_scope: Scope = None
    ) -> Scope:
        scope = ModelScope(
            model.name,
            placement_frame=model.placement_frame,
            canonical_link=model.canonical_link,
        )

        scope.pose = model.pose

        for link in model.links:
            self.convert_link(link, scope)

        for include in model.include:
            self.resolve_include(include, scope)

            # double-check if this is correct
            scope.add_scaffold(include.name, "0 0 0 0 0 0")
            scope.declare_link(
                SimplePose(scope.default_frame, include.name, "0 0 0 0 0 0")
            )

        for nested_model in model.models:
            self.convert_model(nested_model, parent_scope=scope)

            # double-check if this is correct
            scope.add_scaffold(nested_model.name, "0 0 0 0 0 0")
            scope.declare_link(
                SimplePose(scope.default_frame, nested_model.name, "0 0 0 0 0 0")
            )

        for frame in model.frames:
            scope.declare_frame(frame.name)
            scope.add_scaffold(frame.name, frame.pose.value, frame.pose.relative_to)
            if frame.attached_to == "__model__":
                scope.declare_link(DynamicPose(scope.canonical_link, frame.name))
            else:
                scope.declare_link(DynamicPose(frame.attached_to, frame.name))

        for joint in model.joints:
            self.convert_joint(joint, scope)

        if parent_scope is not None:
            parent_scope.add_subscope(scope)
            parent_scope.declare_frame(model.name)
            child = scope.name + "::" + scope.placement_frame

            if model.pose.relative_to is None:
                model.pose.relative_to = model.name

            parent_scope.add_scaffold(child, model.pose.value, model.pose.relative_to)

        if scope.canonical_link is None:
            raise sdformat.ParseError(
                f"Unable to determine canonical link for model '{model.name}'"
            )

        return scope

    def resolve_include(self, include: GenericInclude, scope: Scope) -> None:
        subscope = self._resolve_include(include.uri)

        if include.name is None:
            include.name = subscope.name

        name = include.name
        subscope.name = include.name
        scope.add_subscope(subscope)
        scope.declare_frame(name)

        if isinstance(scope, ModelScope):
            if scope.canonical_link is None:
                scope.canonical_link = name

        if include.placement_frame is not None:
            placement_frame = include.placement_frame
        else:
            placement_frame = subscope.placement_frame
            # TODO: deal with absend //include/pose
            # not quite sure how yet.

        if include.pose is None:
            include.pose = subscope.pose

        if include.pose.relative_to is None:
            include.pose.relative_to = name

        placement_frame = subscope.name + "::" + placement_frame

        scope.add_scaffold(
            placement_frame, include.pose.value, include.pose.relative_to
        )

        child = subscope.name + "::" + subscope.canonical_link
        scope.declare_link(DynamicPose(name, child))

    def convert_world(self, world: GenericWorld) -> Scope:
        world_scope = WorldScope(world.name)
        for include in world.includes:
            self.resolve_include(include, world_scope)

            # double-check if this is correct
            world_scope.add_scaffold(include.name, "0 0 0 0 0 0")
            world_scope.declare_link(SimplePose("world", include.name, "0 0 0 0 0 0"))

        # TODO: GUI
        # TODO: scene

        for light in world.lights:
            self.convert_light(light, scope=world_scope)

            # not sure if this is valid
            world_scope.declare_link(DynamicPose("world", light.name))

        for frame in world.frames:
            world_scope.declare_frame(frame.name)
            world_scope.add_scaffold(
                frame.name, frame.pose.value, frame.pose.relative_to
            )

            if frame.attached_to is None:
                frame.attached_to = "world"
            elif frame.attached_to == "":
                frame.attached_to = "world"

            world_scope.declare_link(DynamicPose(frame.attached_to, frame.name))

        for model in world.models:
            self.convert_model(model, parent_scope=world_scope)

            # double-check if this is correct
            world_scope.add_scaffold(model.name, "0 0 0 0 0 0")
            world_scope.declare_link(SimplePose("world", model.name, "0 0 0 0 0 0"))

            link_name = model.name + "::" + model.canonical_link
            world_scope.declare_link(DynamicPose(model.name, link_name))

        # TODO: actor
        # TODO: road
        # TODO: spherical coords
        # TODO: state
        # TODO: population

        return world_scope


class TransformFactory:
    """A factory that turns SDF of different versions into generic graphs
    that can then be assembled into transform graphs"""

    converters: Dict[str, Callable[[Any], Union[Scope, List[Scope]]]] = dict()

    def __call__(self, sdf: str, *, root_uri: str = None) -> Scope:
        version = sdformat.get_version(sdf)
        if version not in self.converters.keys():
            # lazy loading of SDF bindings
            converter_module = _converter_roots[version]
            if converter_module is None:
                raise NotImplementedError(f"SDFormat v{version} is not supported yet.")
            mod = importlib.import_module(converter_module, __name__)
            self.converters[version] = mod.Converter

        return self.converters[version](root_uri=root_uri)(sdf)


# a singleton
transform_factory = TransformFactory()



from math import tan
import numpy as np
from numpy.typing import ArrayLike

from ..transform.projections import PerspectiveProjection
from ..transform.affine import Translation
from ..transform.utils3d import FrustumProjection
from ..transform.base import Link


class FrustumProjection(FrustumProjection):
    """Frustum based intrinsic camera transformation

    This links behavior is identical to
    :class:`skbot.transform.FrustumProjection`; however, here the camera points
    along the x-axis instead of the z-axis. This is done to match the behavior
    of ignition, which applies the robotic coordinate frame standard instead of
    the computer vision standard here.

    Parameters
    ----------
    hfov : float
        The angle of the viewing frustum in radians. It is assumed to be less than
        pi (180°).
    image_shape : ArrayLike
        The shape (height, width) of the image plane in pixels.

    See Also
    --------
    :class:`skbot.transform.FrustumProjection`

    Notes
    -----
    This function assumes that ``hfov`` is less than pi (180°).

    Points outside the viewing frustum will still be projected; however, their values lie
    outside the applicable pixel range.

    """

    def __init__(self, hfov: float, image_shape: ArrayLike) -> None:
        Link.__init__(self, 3, 2)

        image_shape = np.asarray(image_shape)

        aspect_ratio = image_shape[1] / image_shape[0]
        amounts = np.array(
            [[1 / (tan(hfov / 2)), 0, 0], [aspect_ratio * 1 / (tan(hfov / 2)), 0, 0]]
        )
        directions = np.array([[0, 2 / image_shape[0], 0], [0, 0, 2 / image_shape[1]]])

        self.proj = PerspectiveProjection(directions, amounts, axis=-1)
        self.tf = Translation(image_shape / 2)



""" Generic Bindings for SDFormat

The classes below are a poor man's version of the data bindings in
sdformat.bindings. They only implement a subset of the full SDFormat spec
relevant to frame graphs, but are independent of any specific version.

The idea is that a converter for a specific SDF version takes SDF objects of that
version and converts them into the generic representation below. The generic
objects are then used to construct the actual frame graph.
"""


from typing import List, Tuple


class GenericPose:
    def __init__(self, *, value: str = "0 0 0 0 0 0", relative_to: str = None) -> None:
        self.value = value
        self.relative_to = relative_to


class PoseBearing:
    def __init__(self, *, pose: GenericPose = None) -> None:
        self.pose = pose
        if self.pose is None:
            self.pose = GenericPose()


class NamedPoseBearing(PoseBearing):
    def __init__(self, *, name: str, pose: GenericPose = None) -> None:
        super().__init__(pose=pose)
        self.name = name


class GenericSensor(PoseBearing):
    def __init__(
        self,
        *,
        name: str,
        type: str,
        pose: GenericPose = None,
        camera: "Camera" = None,
        frames: List["GenericFrame"] = None,
    ) -> None:
        super().__init__(pose=pose)
        self.type = type
        self.camera = camera
        self.name = name
        self.frames = frames

        if frames is None:
            self.frames = list()

    class Camera(PoseBearing):
        def __init__(
            self,
            *,
            name: str,
            pose: GenericPose = None,
            horizontal_fov: float = 1.047,
            image: "Image" = None,
            frames: "GenericFrame" = None,
        ) -> None:
            super().__init__(pose=pose)
            self.horizontal_fov = horizontal_fov
            self.image = image
            self.name = name
            self.frames = frames

            if self.frames is None:
                self.frames = list()

            if self.image is None:
                self.image = GenericSensor.Camera.Image()

        class Image:
            def __init__(
                self, *, width: int = 320, height: int = 240, format: str = "R8G8B8"
            ) -> None:
                self.width = width
                self.height = height
                self.format = format


class GenericJoint(PoseBearing):
    def __init__(
        self,
        *,
        name: str,
        kind: str,
        parent: str,
        child: str,
        axis: "Axis" = None,
        pose: GenericPose = None,
        sensor: List[GenericSensor] = None,
        frames: List["GenericFrame"] = None,
    ) -> None:
        super().__init__(pose=pose)
        self.name = name
        self.type = kind
        self.parent = parent
        self.child = child
        self.axis = axis
        self.sensor = sensor
        self.frames = frames

        if axis is None:
            self.axis = GenericJoint.Axis()

        if self.axis.xyz.expressed_in is None:
            self.axis.xyz.expressed_in = self.child

        if frames is None:
            self.frames = list()

    class Axis:
        def __init__(self) -> None:
            self.xyz = GenericJoint.Axis.Xyz()
            self.limit = GenericJoint.Axis.Limit()

        class Xyz:
            def __init__(self, *, value: str = "0 0 1", expressed_in=None) -> None:
                self.value = value
                self.expressed_in = expressed_in

        class Limit:
            def __init__(
                self,
                *,
                lower: float = -1e17,
                upper: float = 1e17,
                effort: float = -1,
                velocity: float = -1,
                stiffness: float = 1e9,
                dissipation: float = 1,
            ) -> None:
                self.lower = lower
                self.upper = upper
                self.effort = effort
                self.velocity = velocity
                self.stiffness = stiffness
                self.dissipation = dissipation


class GenericLink(PoseBearing):
    def __init__(
        self,
        *,
        name: str,
        pose: GenericPose = None,
        must_be_base_link: bool = False,
        inertial: "GenericLink.Inertial" = None,
        collisions: List[NamedPoseBearing],
        visuals: List[NamedPoseBearing],
        projector: NamedPoseBearing = None,
        audio_source_poses: List[GenericPose],
        sensors: List[GenericSensor],
        lights: List["GenericLight"] = None,
        frames: List["GenericFrame"] = None,
    ) -> None:
        super().__init__(pose=pose)
        self.must_be_base_link = must_be_base_link
        self.name = name
        self.inertial = inertial
        self.projector = projector
        self.collision = collisions
        self.visual = visuals
        self.sensors = sensors
        self.lights = lights
        self.audio_sources = [PoseBearing(pose=p) for p in audio_source_poses]
        self.frames = frames

        if frames is None:
            self.frames = list()

        if lights is None:
            self.lights = list()

    class Inertial(PoseBearing):
        def __init__(
            self, *, pose: GenericPose = None, frames: List["GenericFrame"] = None
        ) -> None:
            super().__init__(pose=pose)
            self.frames = frames

            if self.frames is None:
                self.frames = list()


class GenericLight(NamedPoseBearing):
    def __init__(
        self,
        *,
        name: str,
        pose: GenericPose = None,
        frames: List["GenericFrame"] = None,
    ) -> None:
        super().__init__(name=name, pose=pose)

        self.frames = frames

        if frames is None:
            self.frames = list()


class GenericInclude(NamedPoseBearing):
    def __init__(
        self,
        *,
        name: str,
        pose: GenericPose = None,
        placement_frame: str = None,
        uri: str,
    ) -> None:
        super().__init__(name=name, pose=pose)
        self.placement_frame = placement_frame
        self.uri = uri
        if pose is None:
            # here pose is None means no overwrite
            self.pose = None


class GenericFrame(NamedPoseBearing):
    def __init__(
        self, *, name: str, pose: GenericPose, attached_to: str = None
    ) -> None:
        super().__init__(name=name, pose=pose)
        self.attached_to = attached_to


class GenericModel(NamedPoseBearing):
    def __init__(
        self,
        *,
        name: str,
        pose: GenericPose = None,
        placement_frame: str = None,
        canonical_link: str = None,
        links: List[GenericLink],
        include: List[GenericInclude],
        models: List["GenericModel"],
        frames: List[GenericFrame],
        joints: List[GenericJoint],
    ) -> None:
        super().__init__(name=name, pose=pose)
        self.placement_frame = placement_frame
        self.canonical_link = canonical_link
        self.links = links
        self.include = include
        self.models = models
        self.frames = frames
        self.joints = joints

        if self.canonical_link is None:
            if len(links) > 0:
                self.canonical_link = links[0].name
            elif len(include) > 0:
                self.canonical_link = include[0].name
            elif len(models) > 0:
                self.canonical_link = models[0].name

        if self.placement_frame is None:
            self.placement_frame = "__model__"

        for frame in self.frames:
            if frame.attached_to is None:
                frame.attached_to = self.placement_frame
            elif frame.attached_to == "":
                frame.attached_to = self.placement_frame


class GenericWorld:
    def __init__(
        self,
        *,
        name: str,
        includes: List[GenericInclude],
        lights: List[GenericLight],
        frames: List[GenericFrame],
        models: List[GenericModel],
        population: List["GenericPopulation"],
    ) -> None:
        self.name = name
        self.includes = includes
        self.lights = lights
        self.frames = frames
        self.models = models
        self.population = population

    class GenericPopulation(NamedPoseBearing):
        def __init__(
            self,
            *,
            name: str,
            pose: GenericPose = None,
            model_count: int = 1,
            distribution: "GenericDistribution",
            box: "GenericBox" = None,
            cylinder: "GenericCylinder" = None,
            model: GenericModel,
            frames: List[GenericFrame] = None,
        ) -> None:
            super().__init__(name=name, pose=pose)
            self.model_count = model_count
            self.distribution = distribution
            self.box = box
            self.cylinder = cylinder
            self.model = model
            self.frames = frames

            if frames is None:
                self.frames = list()

        class GenericDistribution:
            def __init__(
                self,
                *,
                kind: str = "random",
                rows: int = 1,
                cols: int = 1,
                step: Tuple[float] = (0.5, 0.5, 0),
            ):
                self.type = kind
                self.rows = rows
                self.cols = cols
                self.step = step

        class GenericBox:
            def __init__(self, *, size: Tuple[float] = (1, 1, 1)) -> None:
                self.size = size

        class GenericCylinder:
            def __init__(self, *, radius: float = 1, length: float = 1) -> None:
                self.radius = radius
                self.length = length



from typing import Dict, List, Union, Tuple
import numpy as np

from .... import transform as tf
from .. import sdformat


class SdfLink:
    """Base class for link factory."""

    def __init__(
        self, parent: Union[str, tf.Frame], child: Union[str, tf.Frame]
    ) -> None:
        self.parent: Union[str, tf.Frame] = parent
        self.child: Union[str, tf.Frame] = child

    def to_transform_link(self, scope: "Scope") -> tf.Link:
        raise NotImplementedError()


class ScaffoldPose(SdfLink):
    """Pose instance for constructing the pose scaffold."""

    def __init__(self, parent: str, child: Union[str, tf.Frame], pose: str) -> None:
        super().__init__(parent, child)
        self.pose = np.array(pose.split(), dtype=float)

    def to_transform_link(self, scope: "Scope") -> tf.Link:
        return tf.CompundLink(
            [tf.EulerRotation("xyz", self.pose[3:]), tf.Translation(self.pose[:3])]
        )


class Scope:
    """A scope within SDFormat.

    This class does most of the heavy lifting of sdformat.to_frame_graph. It
    collects frames that are (explicitly or implicitly) declared within a SDF
    scope and holds a scope's sub-scopes (if any).

    SDF positions objects using a <pose> element, which can be relative to
    another element or relative to the scope's implicit root frame. This class
    first collects all such definitions and uses them to construct a static
    graph referred to as scaffolding (self.build_scaffolding).

    Once the (static) position of all objects can be computed using the
    scaffolding this class then constructs a dynamic pose graph that reflects
    the actual dynamic links between frames (self.resolve_links). The root
    scope's dynamic frame graph is then returned by to_frame_graph.

    """

    def __init__(self, name, *, parent: "Scope" = None) -> None:
        self.nested_scopes: Dict[str, "Scope"] = dict()

        self.frames: Dict[str, tf.Frame] = dict()
        self.links: List[SdfLink] = list()

        self.scaffold_frames: Dict[str, tf.Frame] = dict()
        self.scaffold_links: List[ScaffoldPose] = list()

        # instantiated downstream
        self.default_frame: tf.Frame = None

        # might be able to remove this
        self.name = name

        self.parent = parent
        self.placement_frame: str = None

    def declare_frame(self, name: str) -> None:
        """Add a frame to the scaffolding and final frame graph.

        The declared frame can be found by dynamic links using the given name.

        """

        if name in self.frames.keys():
            raise sdformat.ParseError(f"Frame '{name}' already declared.")

        if name == "world":
            raise sdformat.ParseError(
                "Can not create a frame named 'world' (reserved name)."
            )

        self.frames[name] = tf.Frame(3, name=name)
        self.scaffold_frames[name] = tf.Frame(3, name=name)

    def add_scaffold(self, frame_name: str, pose: str, relative_to: str = None) -> None:
        """Add a pose to the scaffolding frame graph."""

        parent = self.default_frame.name
        if relative_to is not None and not relative_to == "":
            parent = relative_to

        self.scaffold_links.append(ScaffoldPose(parent, frame_name, pose))

    def declare_link(self, link: SdfLink) -> None:
        """Add a link to the final/dynamic frame graph."""
        self.links.append(link)

    def get(self, name: str, scaffolding: bool) -> tf.Frame:
        """Find the frame from a (namespaced) SDFormat name."""

        if "::" in name:
            scope, subscope_name = name.split("::", 1)

            try:
                frame = self.nested_scopes[scope].get(subscope_name, scaffolding)
            except sdformat.ParseError:
                raise sdformat.ParseError(
                    f"No frame named '{name}' in scope '{self.name}'"
                ) from None
        else:
            storage = self.frames
            if scaffolding:
                storage = self.scaffold_frames

            try:
                frame = storage[name]
            except KeyError:
                raise sdformat.ParseError(
                    f"No frame named '{name}' in scope '{self.name}'"
                ) from None

        return frame

    def build_scaffolding(self) -> None:
        """Construct the static frame graph.

        Connect declared scaffolding frames with declared scaffolding links.
        """
        for el in self.scaffold_links:
            tf_link = el.to_transform_link(self)

            parent = self.get(el.parent, scaffolding=True)

            child = el.child
            if isinstance(child, str):
                child = self.get(child, scaffolding=True)

            tf_link(parent, child)

        for scope in self.nested_scopes.values():
            scope.build_scaffolding()

    def resolve_links(self, *, shape: Tuple[int], axis: int) -> None:
        """Construct the dynamic frame graph.

        The success of this function depends on build_scaffolding. Links in the
        dynamic frame graph may have to be computed using the scaffolding and hence
        it needs to have been build prior to the execution of this function.

        """
        for el in self.links:
            if isinstance(el.parent, str):
                parent = self.get(el.parent, scaffolding=False)
            else:
                parent = el.parent

            if isinstance(el.child, str):
                child = self.get(el.child, scaffolding=False)
            else:
                child = el.child

            try:
                tf_link = el.to_transform_link(self, shape, axis)
            except RuntimeError:
                raise sdformat.ParseError(
                    f"Unable to express pose of '{child.name}' in frame '{parent.name}'."
                )
            tf_link(parent, child)

        for scope in self.nested_scopes.values():
            scope.resolve_links(shape=shape, axis=axis)

    def add_subscope(self, nested_scope: "Scope") -> None:
        """Add a scope nested within this scope."""

        if nested_scope.name in self.nested_scopes.keys():
            raise sdformat.ParseError(
                f"Nested Scope '{nested_scope.name}' already defined."
            )

        self.nested_scopes[nested_scope.name] = nested_scope
        nested_scope.parent = self


class ModelScope(Scope):
    """Specialization for scope declared by <model> elements."""

    def __init__(
        self,
        name,
        *,
        parent: "Scope" = None,
        placement_frame: str = None,
        canonical_link: str = None,
    ) -> None:
        super().__init__(name, parent=parent)

        self.placement_frame = placement_frame

        self.default_frame = tf.Frame(3, name="__model__")
        self.scaffold_frames["__model__"] = self.default_frame

        self.canonical_link = canonical_link

        self.pose = None

    def get(self, name: str, scaffolding: bool) -> tf.Frame:
        """Find the frame from a (namespaced) SDFormat name."""

        if name == "world":
            return self.parent.get(name, scaffolding)

        return super().get(name, scaffolding)


class WorldScope(Scope):
    """Specialization for scopes declared by <world> elements."""

    def __init__(self, name) -> None:
        super().__init__(name, parent=None)

        self.default_frame = tf.Frame(3, name="world")
        self.scaffold_frames["world"] = self.default_frame
        self.frames["world"] = tf.Frame(3, name=name)

    def get(self, name: str, scaffolding: bool) -> tf.Frame:
        """Find the frame from a (namespaced) SDFormat name"""

        if name == "world":
            if scaffolding:
                return self.default_frame
            else:
                return self.frames["world"]

        return super().get(name, scaffolding)


class LightScope(Scope):
    """Specialization for scopes declared by <light> elements."""

    def __init__(self, name) -> None:
        super().__init__(name, parent=None)

        self.default_frame = tf.Frame(3, name="__light__")
        self.scaffold_frames["__light__"] = self.default_frame



from typing import Union, List, Tuple

from ... import transform as tf
from ._transform_factory.scopes import LightScope, ModelScope, Scope, WorldScope
from ._transform_factory.factory import transform_factory


def to_frame_graph(
    sdf: str, *, unwrap: bool = True, shape: Tuple[int] = (3,), axis: int = -1
) -> Union[tf.Frame, List[tf.Frame]]:
    """Create a frame graph from a sdformat string.

    .. versionadded:: 0.6.0
        This function has been added to the library.

    Parameters
    ----------
    sdf: str
        A string containing SDFormat XML.
    unwrap : bool
        If True (default) and the sdf only contains a single light, model, or
        world element return that element's frame. If the sdf contains multiple
        lights, models, or worlds a list of root frames is returned. If False,
        always return a list of frames.
    shape : tuple
        A tuple describing the shape of elements that the resulting graph should
        transform. This can be used to add batch dimensions to the graph, for
        example to perform vectorized computation on multiple instances of the
        same world in different states, or for batched coordinate
        transformation. Defaults to (3,), which is the shape of a single 3D
        vector in euclidian space.
    axis : int
        The axis along which elements are stored. The axis must have length 3
        (since SDFormat describes 3 dimensional worlds), and all other axis are
        considered batch dimensions. Defaults to -1.

    Returns
    -------
    frame_graph : Union[Frame, List[Frame]]
        A :class:`skbot.transform.Frame` or list of Frames depending on the value of
        ``unwrap`` and the number of elements in the SDF's root element.

    See Also
    --------
    :mod:`skbot.transform`

    Notes
    -----
    Frames inside the graph are named after the frames defined by the SDF. You can
    retrieve them by searching for them using :func:`skbot.transform.Frame.find_frame`.

    Joins are implicit within the frame graph. The joint frame that is attached
    to the child frame is named after the joint (<joint_name>), and the
    (implicit) joint frame attached to the parent is named
    (<joint_name>_parent). The link between the two frames can be retrieved via
    :func:`skbot.transform.Frame.transform_chain`. For example if there is a
    joint named "robot_joint0" its link can be retrieved using::

        child_frame = frame_graph.find_frame(".../robot_joint0")
        parent_frame = frame_graph.find_frame(".../robot_joint0_parent")
        link = child_frame.transform_chain(child_frame)[0]
    """

    scope_list: List[Scope] = transform_factory(sdf)
    frame_list: List[tf.Frame] = list()
    for scope in scope_list:
        scope.build_scaffolding()
        scope.resolve_links(shape=shape, axis=axis)

        if isinstance(scope, WorldScope):
            frame_list.append(scope.get("world", scaffolding=False))
        elif isinstance(scope, ModelScope):
            frame = scope.get(scope.canonical_link, scaffolding=False)
            frame_list.append(frame)
        else:
            scope: LightScope
            frame = scope.get(scope.name, scaffolding=False)
            frame_list.append(frame)

    if unwrap and len(frame_list) == 1:
        return frame_list[0]
    else:
        return frame_list
